#!/usr/bin/python3
# pylint: disable=wrong-import-position,too-many-locals,unused-argument,wrong-import-order
#
# Copyright (C) 2017 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import gi
import os
import requests
import time
import sys
import glob
import json
from termcolor import colored

gi.require_version('Fwupd', '2.0')

from gi.repository import Fwupd
from gi.repository import Gio
from gi.repository import GLib


def _get_cache_file(fn):
    cachedir = os.path.expanduser('~/.cache/fwupdmgr')
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)
    cachefn = os.path.join(cachedir, fn)
    if not os.path.exists(cachefn):
        url = 'https://fwupd.org/downloads/' + fn
        print("Downloading", url)
        r = requests.get(url)
        r.raise_for_status()
        f = open(cachefn, 'wb')
        f.write(r.content)
        f.close()
    return cachefn


class DeviceTest:
    def __init__(self, obj):
        self.client = Fwupd.Client.new()
        self.name = obj.get('name', 'Unknown')
        self.guids = obj.get('guids', [])
        self.releases = obj.get('releases', [])
        self.has_runtime = obj.get('runtime', True)
        self.interactive = obj.get('interactive', False)
        self.disabled = obj.get('disabled', False)
        self.protocol = obj.get('protocol', None)

    def _info(self, msg):
        print(colored('[INFO]'.ljust(10), 'blue'), msg)

    def _warn(self, msg):
        print(colored('[WARN]'.ljust(10), 'yellow'), msg)

    def _failed(self, msg):
        print(colored('[FAILED]'.ljust(10), 'red'), msg)

    def _success(self, msg):
        print(colored('[SUCCESS]'.ljust(10), 'green'), msg)

    def _get_by_device_guids(self):
        cancellable = Gio.Cancellable.new()
        for d in self.client.get_devices(cancellable):
            for guid in self.guids:
                if d.has_guid(guid):
                    if self.protocol and self.protocol != d.get_protocol():
                        continue
                    return d
        return None

    def run(self):

        print('Running test on {}'.format(self.name))
        dev = self._get_by_device_guids()
        if not dev:
            self._warn('no {} attached'.format(self.name))
            return

        self._info('Current version {}'.format(dev.get_version()))

        # apply each file
        for obj in self.releases:
            ver = obj.get('version')
            fn = obj.get('file')
            repeat = obj.get('repeat', 1)
            try:
                fn_cache = _get_cache_file(fn)
            except requests.exceptions.HTTPError as e:
                self._failed('Failed to download: {}'.format(str(e)))
                return

            # some hardware updates more than one partition with the same firmware
            for cnt in range(0, repeat):
                if dev.get_version() == ver:
                    flags = Fwupd.InstallFlags.ALLOW_REINSTALL
                    self._info('Reinstalling version {}'.format(ver))
                else:
                    flags = Fwupd.InstallFlags.ALLOW_OLDER
                    self._info('Installing version {}'.format(ver))
                cancellable = Gio.Cancellable.new()

                try:
                    self.client.install(dev.get_id(), fn_cache, flags, cancellable)
                except GLib.Error as e:
                    if str(e).find('no HWIDs matched') != -1:
                        self._info('Skipping as {}'.format(e))
                        continue
                    self._failed('Could not install: {}'.format(e))
                    return

                # verify version
                if self.has_runtime:
                    dev = self._get_by_device_guids()
                    if not dev:
                        self._failed('Device did not come back: ' + self.name)
                        return
                    if not dev.get_version():
                        self._failed('No version set after flash for: ' + self.name)
                        return
                    if cnt == repeat - 1 and dev.get_version() != ver:
                        self._failed('Got: ' + dev.get_version() + ', expected: ' + ver)
                        return
                    self._success('Installed {}'.format(dev.get_version()))
                else:
                    self._success('Assumed success (no runtime)')

            # wait for device to settle?
            time.sleep(2)


if __name__ == '__main__':

    # get manifests to parse
    device_fns = []
    if len(sys.argv) == 1:
        device_fns.extend(glob.glob('devices/*.json'))
    else:
        for fn in sys.argv[1:]:
            device_fns.append(fn)

    # run each test
    for fn in sorted(device_fns):
        print('{}:'.format(fn))
        with open(fn, 'r') as f:
            try:
                obj = json.load(f)
            except json.decoder.JSONDecodeError as e:
                print('Failed to parse {}: {}'.format(fn, e))
                continue
        t = DeviceTest(obj)
        if t.disabled:
            continue
        if t.interactive and len(device_fns) > 1:
            continue
        t.run()
    sys.exit(0)



#!/usr/bin/python3
#
# Copyright (C) 2020 Dell Inc.
#
# SPDX-License-Identifier: LGPL-2.1+
#
import os
import argparse
import xml.etree.ElementTree as etree

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("version", help="Generate news for release")
    args = parser.parse_args()

    tree = etree.parse(os.path.join("data", "org.freedesktop.fwupd.metainfo.xml"))
    root = tree.getroot()
    for release in root.iter("release"):
        if "version" not in release.attrib:
            continue
        if release.attrib["version"] != args.version:
            continue
        description = release.find("description")
        str = etree.tostring(description, encoding="unicode", method="text")
        print(str.strip())



#!/usr/bin/python3
#
# Copyright (C) 2017 Max Ehrlich maxehr@gmail.com
#
# SPDX-License-Identifier: LGPL-2.1+
#

import argparse
import subprocess
import contextlib
import os
import shutil
import tempfile
import time


@contextlib.contextmanager
def cd(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(prev_cwd)


firmware_metainfo_template = """
<?xml version="1.0" encoding="UTF-8"?>
<component type="firmware">
  <id>org.{developer_name}.guid{firmware_id}</id>
  <name>{firmware_name}</name>
  <summary>{firmware_summary}</summary>
  <description>
    {firmware_description}
  </description>
  <provides>
    <firmware type="flashed">{device_guid}</firmware>
  </provides>
  <url type="homepage">{firmware_homepage}</url>
  <metadata_license>CC0-1.0</metadata_license>
  <project_license>proprietary</project_license>
  <updatecontact>{contact_info}</updatecontact>
  <developer_name>{developer_name}</developer_name>
  <releases>
    <release version="{release_version}" timestamp="{timestamp}">
      <description>
        {release_description}
      </description>
    </release>
  </releases>
</component>
"""


def make_firmware_metainfo(firmware_info, dst):
    local_info = vars(firmware_info)
    local_info["firmware_id"] = local_info["device_guid"][0:8]
    firmware_metainfo = firmware_metainfo_template.format(
        **local_info, timestamp=time.time()
    )

    with open(os.path.join(dst, 'firmware.metainfo.xml'), 'w') as f:
        f.write(firmware_metainfo)


def extract_exe(exe, dst):
    command = ['7z', 'x', '-o{}'.format(dst), exe]
    subprocess.check_call(command, stdout=subprocess.DEVNULL)


def get_firmware_bin(root, bin_path, dst):
    with cd(root):
        shutil.copy(bin_path, os.path.join(dst, 'firmware.bin'))


def create_firmware_cab(exe, folder):
    with cd(folder):
        if os.name == "nt":
            directive = os.path.join(folder, "directive")
            with open(directive, 'w') as wfd:
                wfd.write('.OPTION EXPLICIT\r\n')
                wfd.write('.Set CabinetNameTemplate=firmware.cab\r\n')
                wfd.write('.Set DiskDirectory1=.\r\n')
                wfd.write('firmware.bin\r\n')
                wfd.write('firmware.metainfo.xml\r\n')
            command = ['makecab.exe', '/f', directive]
        else:
            command = [
                'gcab',
                '--create',
                'firmware.cab',
                'firmware.bin',
                'firmware.metainfo.xml',
            ]
        subprocess.check_call(command)


def main(args):
    with tempfile.TemporaryDirectory() as dir:
        print('Using temp directory {}'.format(dir))

        if args.exe:
            print('Extracting firmware exe')
            extract_exe(args.exe, dir)

        print('Locating firmware bin')
        get_firmware_bin(dir, args.bin, dir)

        print('Creating metainfo')
        make_firmware_metainfo(args, dir)

        print('Creating cabinet file')
        create_firmware_cab(args, dir)

        print('Done')
        shutil.copy(os.path.join(dir, 'firmware.cab'), args.out)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Create fwupd packaged from windows executables'
    )
    parser.add_argument(
        '--firmware-name',
        help='Name of the firmware package can be customized (e.g. DellTBT)',
        required=True,
    )
    parser.add_argument(
        '--firmware-summary', help='One line description of the firmware package'
    )
    parser.add_argument(
        '--firmware-description', help='Longer description of the firmware package'
    )
    parser.add_argument(
        '--device-guid',
        help='GUID of the device this firmware will run on, this *must* match the output of one of the GUIDs in `fwupdmgr get-devices`',
        required=True,
    )
    parser.add_argument('--firmware-homepage', help='Website for the firmware provider')
    parser.add_argument(
        '--contact-info', help='Email address of the firmware developer'
    )
    parser.add_argument(
        '--developer-name', help='Name of the firmware developer', required=True
    )
    parser.add_argument(
        '--release-version',
        help='Version number of the firmware package',
        required=True,
    )
    parser.add_argument(
        '--release-description', help='Description of the firmware release'
    )
    parser.add_argument(
        '--exe', help='(optional) Executable file to extract firmware from'
    )
    parser.add_argument(
        '--bin',
        help='Path to the .bin file (Relative if inside the executable; Absolute if outside) to use as the firmware image',
        required=True,
    )
    parser.add_argument('--out', help='Output cab file path', required=True)
    args = parser.parse_args()

    main(args)



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright (C) 2021 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#
# pylint: disable=too-many-instance-attributes,no-self-use

import os
import sys
import subprocess
import glob
from typing import Dict, Optional, List, Union

DEFAULT_BUILDDIR = ".ossfuzz"


class Builder:
    def __init__(self) -> None:

        self.cc = self._ensure_environ("CC", "gcc")
        self.cxx = self._ensure_environ("CXX", "g++")
        self.builddir = self._ensure_environ("WORK", os.path.realpath(DEFAULT_BUILDDIR))
        self.installdir = self._ensure_environ(
            "OUT", os.path.realpath(os.path.join(DEFAULT_BUILDDIR, "out"))
        )
        self.srcdir = self._ensure_environ("SRC", os.path.realpath(".."))
        self.ldflags = ["-lpthread", "-lresolv", "-ldl", "-lffi", "-lz"]

        # defined in env
        self.cflags = ["-Wno-deprecated-declarations"]
        if "CFLAGS" in os.environ:
            self.cflags += os.environ["CFLAGS"].split(" ")
        self.cxxflags = []
        if "CXXFLAGS" in os.environ:
            self.cxxflags += os.environ["CXXFLAGS"].split(" ")

        # set up shared / static
        os.environ["PKG_CONFIG"] = "pkg-config --static"
        if "PATH" in os.environ:
            os.environ["PATH"] = "{}:{}".format(
                os.environ["PATH"], os.path.join(self.builddir, "bin")
            )
        else:
            os.environ["PATH"] = os.path.join(self.builddir, "bin")
        os.environ["PKG_CONFIG_PATH"] = os.path.join(self.builddir, "lib", "pkgconfig")

        # writable
        os.makedirs(self.builddir, exist_ok=True)
        os.makedirs(self.installdir, exist_ok=True)

    def _ensure_environ(self, key: str, value: str) -> str:
        """ set the environment unless already set """
        if key not in os.environ:
            os.environ[key] = value
        return os.environ[key]

    def checkout_source(self, name: str, url: str, commit: Optional[str] = None) -> str:
        """ checkout source tree, optionally to a specific commit """
        srcdir_name = os.path.join(self.srcdir, name)
        if os.path.exists(srcdir_name):
            return srcdir_name
        subprocess.run(["git", "clone", url], cwd=self.srcdir, check=True)
        if commit:
            subprocess.run(["git", "checkout", commit], cwd=srcdir_name, check=True)
        return srcdir_name

    def build_meson_project(self, srcdir: str, argv) -> None:
        """ configure and build the meson project """
        srcdir_build = os.path.join(srcdir, DEFAULT_BUILDDIR)
        if not os.path.exists(srcdir_build):
            subprocess.run(
                [
                    "meson",
                    "--prefix",
                    self.builddir,
                    "--libdir",
                    "lib",
                    "--default-library",
                    "static",
                ]
                + argv
                + [DEFAULT_BUILDDIR],
                cwd=srcdir,
                check=True,
            )
            subprocess.run(["ninja", "install"], cwd=srcdir_build, check=True)

    def add_work_includedir(self, value: str) -> None:
        """ add a CFLAG """
        self.cflags.append("-I{}/{}".format(self.builddir, value))

    def add_src_includedir(self, value: str) -> None:
        """ add a CFLAG """
        self.cflags.append("-I{}/{}".format(self.srcdir, value))

    def add_build_ldflag(self, value: str) -> None:
        """ add a LDFLAG """
        self.ldflags.append(os.path.join(self.builddir, value))

    def substitute(self, src: str, replacements: Dict[str, str]) -> str:
        """ map changes """

        dst = os.path.basename(src).replace(".in", "")
        with open(os.path.join(self.srcdir, src), "r") as f:
            blob = f.read()
            for key in replacements:
                blob = blob.replace(key, replacements[key])
            with open(os.path.join(self.builddir, dst), "w") as out:
                out.write(blob)
        return dst

    def compile(self, src: str) -> str:
        """ compile a specific source file """
        argv = [self.cc]
        argv.extend(self.cflags)
        fullsrc = os.path.join(self.srcdir, src)
        if not os.path.exists(fullsrc):
            fullsrc = os.path.join(self.builddir, src)
        dst = os.path.basename(src).replace(".c", ".o")
        argv.extend(
            [
                "-c",
                fullsrc,
                "-o",
                os.path.join(self.builddir, dst),
            ]
        )
        print("building {} into {}".format(src, dst))
        try:
            subprocess.run(argv, cwd=self.srcdir, check=True)
        except subprocess.CalledProcessError as e:
            print(e)
            sys.exit(1)
        return os.path.join(self.builddir, "{}".format(dst))

    def link(self, objs: List[str], dst: str) -> None:
        """ link multiple obects into a binary """
        argv = [self.cxx] + self.cxxflags
        for obj in objs:
            if obj.startswith("-"):
                argv.append(obj)
            else:
                argv.append(os.path.join(self.builddir, obj))
        argv += ["-o", os.path.join(self.installdir, dst)]
        argv += self.ldflags
        print("building {} into {}".format(",".join(objs), dst))
        subprocess.run(argv, cwd=self.srcdir, check=True)

    def write_header(
        self, dst: str, defines: Dict[str, Optional[Union[str, int]]]
    ) -> None:
        """ write a header file """
        dstdir = os.path.join(self.builddir, os.path.dirname(dst))
        os.makedirs(dstdir, exist_ok=True)
        print("writing {}".format(dst))
        with open(os.path.join(dstdir, os.path.basename(dst)), "w") as f:
            for key in defines:
                value = defines[key]
                if value is not None:
                    if isinstance(value, int):
                        f.write("#define {} {}\n".format(key, value))
                    else:
                        f.write('#define {} "{}"\n'.format(key, value))
                else:
                    f.write("#define {}\n".format(key))
        self.add_work_includedir(os.path.dirname(dst))

    def makezip(self, dst: str, globstr: str) -> None:
        """ create a zip file archive from a glob """
        argv = [
            "zip",
            "--junk-paths",
            os.path.join(self.installdir, dst),
        ] + glob.glob(os.path.join(self.srcdir, globstr))
        print("assembling {}".format(dst))
        subprocess.run(argv, cwd=self.srcdir, check=True)

    def grep_meson(self, src: str, token: str = "fuzzing") -> List[str]:
        """ find source files tagged with a specific comment """
        srcs = []
        with open(os.path.join(self.srcdir, src, "meson.build"), "r") as f:
            for line in f.read().split("\n"):
                if line.find(token) == -1:
                    continue
                srcs.append(
                    os.path.join(
                        src,
                        line.strip()
                        .replace("'", "")
                        .replace(",", "")
                        .replace(" ", "")
                        .split("#")[0],
                    )
                )
        return srcs


def _build(bld: Builder) -> None:

    # GLib
    src = bld.checkout_source("glib", url="https://gitlab.gnome.org/GNOME/glib.git")
    bld.build_meson_project(
        src,
        [
            "-Dlibmount=disabled",
            "-Dselinux=disabled",
            "-Dnls=disabled",
            "-Dlibelf=disabled",
            "-Dbsymbolic_functions=false",
            "-Dtests=false",
            "-Dinternal_pcre=true",
        ],
    )
    bld.add_work_includedir("include/glib-2.0")
    bld.add_work_includedir("lib/glib-2.0/include")
    bld.add_build_ldflag("lib/libgio-2.0.a")
    bld.add_build_ldflag("lib/libgmodule-2.0.a")
    bld.add_build_ldflag("lib/libgobject-2.0.a")
    bld.add_build_ldflag("lib/libglib-2.0.a")
    bld.add_build_ldflag("lib/libgthread-2.0.a")

    # JSON-GLib
    src = bld.checkout_source(
        "json-glib",
        url="https://gitlab.gnome.org/GNOME/json-glib.git",
    )
    bld.build_meson_project(
        src,
        [
            "-Dgtk_doc=disabled",
            "-Dtests=false",
            "-Dintrospection=disabled",
        ],
    )
    bld.add_work_includedir("include/json-glib-1.0/json-glib")
    bld.add_work_includedir("include/json-glib-1.0")
    bld.add_build_ldflag("lib/libjson-glib-1.0.a")

    # libxmlb
    src = bld.checkout_source("libxmlb", url="https://github.com/hughsie/libxmlb.git")
    bld.build_meson_project(
        src,
        [
            "-Dgtkdoc=false",
            "-Dintrospection=false",
            "-Dtests=false",
        ],
    )
    bld.add_work_includedir("include/libxmlb-2")
    bld.add_work_includedir("include/libxmlb-2/libxmlb")
    bld.add_build_ldflag("lib/libxmlb.a")

    # write required headers
    bld.write_header("libfwupd/fwupd-version.h", {})
    bld.write_header(
        "config.h",
        {
            "FWUPD_DATADIR": "/tmp",
            "FWUPD_LOCALSTATEDIR": "/tmp",
            "FWUPD_PLUGINDIR": "/tmp",
            "FWUPD_SYSCONFDIR": "/tmp",
            "HAVE_REALPATH": None,
            "PACKAGE_NAME": "fwupd",
            "PACKAGE_VERSION": "0.0.0",
        },
    )

    # libfwupd + libfwupdplugin
    built_objs: List[str] = []
    bld.add_src_includedir("fwupd")
    for path in ["fwupd/libfwupd", "fwupd/libfwupdplugin"]:
        bld.add_src_includedir(path)
        for src in bld.grep_meson(path):
            built_objs.append(bld.compile(src))

    # dummy binary entrypoint
    if "LIB_FUZZING_ENGINE" in os.environ:
        built_objs.append(os.environ["LIB_FUZZING_ENGINE"])
    else:
        built_objs.append(bld.compile("fwupd/libfwupdplugin/fu-fuzzer-main.c"))

    # built in formats
    for fuzzer in ["dfuse", "fmap", "ihex", "srec"]:
        src = bld.substitute(
            "fwupd/libfwupdplugin/fu-fuzzer-firmware.c.in",
            {
                "@FIRMWARENEW@": "fu_{}_firmware_new".format(fuzzer),
                "@INCLUDE@": "libfwupdplugin/fu-{}-firmware.h".format(fuzzer),
            },
        )
        bld.link([bld.compile(src)] + built_objs, "{}_fuzzer".format(fuzzer))
        bld.makezip(
            "{}_fuzzer_seed_corpus.zip".format(fuzzer),
            "fwupd/src/fuzzing/firmware/{}*".format(fuzzer),
        )

    # plugins
    for srcdir, fuzzer, globstr in [
        ("bcm57xx", "bcm57xx", "bcm57xx*"),
        ("ccgx", "ccgx", "ccgx*.cyacd"),
        ("ccgx", "ccgx-dmc", "ccgx-dmc*.bin"),
        ("cros-ec", "cros-ec", "cros-ec*"),
        ("ebitdo", "ebitdo", "ebitdo*"),
        ("hailuck", "hailuck-kbd", "ihex*"),
        ("pixart-rf", "pxi", "pixart*"),
        ("solokey", "solokey", "solokey*"),
        ("synaptics-prometheus", "synaprom", "synaprom*"),
        ("synaptics-rmi", "synaptics-rmi", "synaptics-rmi*"),
        ("synaptics-mst", "synaptics-mst", "synaptics-mst*"),
        ("wacom-usb", "wac", "wacom*"),
    ]:
        fuzz_objs = []
        for obj in bld.grep_meson("fwupd/plugins/{}".format(srcdir)):
            fuzz_objs.append(bld.compile(obj))
        src = bld.substitute(
            "fwupd/libfwupdplugin/fu-fuzzer-firmware.c.in",
            {
                "@FIRMWARENEW@": "fu_{}_firmware_new".format(fuzzer.replace("-", "_")),
                "@INCLUDE@": "plugins/{}/fu-{}-firmware.h".format(srcdir, fuzzer),
            },
        )
        fuzz_objs.append(bld.compile(src))
        bld.link(fuzz_objs + built_objs, "{}_fuzzer".format(fuzzer))
        bld.makezip(
            "{}_fuzzer_seed_corpus.zip".format(fuzzer),
            "fwupd/src/fuzzing/firmware/{}".format(globstr),
        )


if __name__ == "__main__":
    _builder = Builder()
    _build(_builder)
    sys.exit(0)
